import pytest
import random

from .dummy import Dummy
from .integer import Integer
from .string import String
from .util import _test_valid_args, _test_invalid_args, Value, Values


class Array(Dummy):
    def __init__(self, datatype, min_items=None, max_items=None, unique_items=False):
        super().__init__()
        if min_items is not None and min_items < 0:
            raise ValueError("min_items must be creater than 0, not {}".format(mi_items))
        if max_items is not None and max_items < 0:
            raise ValueError("maxItems must be creater than 0, not {}".format(max_items))
        if min_items is not None and max_items is not None and min_items > max_items:
            raise ValueError("min_items ({}) cannot be larger than maxItems ({}).").format(min_items, max_items)
        
        self.datatype = datatype
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items
    
    def valid(self, value, raise_error=False):
        if self.min_items and len(value) < self.min_items:
            if raise_error:
                raise ValueError("{} < {}".format(len(value), self.min_items))
            return False
        if self.max_items and len(value) > self.max_items:
            if raise_error:
                raise ValueError("{} > {}".format(len(value), self.max_items))
            return False
        if self.unique_items:
            uniques = set([str(x) for x in value])
            if len(value) != len(uniques):
                if raise_error:
                    raise ValueError("{} has duplicate elements".format(value))
                return False
        for v in value:
            if not self.datatype.valid(v, raise_error):
                return False
        return True
    
    def _valids(self):
        length = random.randint(*self._valid_length_range())
        return Value(self._valid_list(length))
    
    def _valid_list(self, length):
        if not self.unique_items:
            return [self.datatype.valids(1)[0] for _ in range(length)]
        else:
            valids = []
            duplicates = 0
            while len(valids) < length:
                v = self.datatype.valids(1)[0]
                if v not in valids:
                    valids.append(v)
                else:
                    duplicates += 1
                    if duplicates > 100 * length:
                        raise ValueError("Generated {} duplicate items. ".format(duplicates) +
                                         "Current list size: {} / {}.".format(len(valids), length) +
                                         "Does the datatype generate enough unique values?")
            return valids
    
    def _invalids(self):
        invalids = []
        # too short
        if self.min_items and self.min_items > 0:
            invalids.append(self._valid_list(self.min_items - 1))
        # too long
        if self.max_items:
            invalids.append(self._valid_list(self.max_items + 1))
        # invalid item
        if self._valid_length_range()[1] > 0:
            inv = self.datatype.invalids()
            if len(inv) > 0:
                invalids.append(self._valid_list(self._valid_length_range()[1]))
                invalids[-1][-1] = inv[0]
        # duplicate items
        if self.unique_items:
            if self._valid_length_range()[1] > 1:
                invalids.append(self._valid_list(self._valid_length_range()[1]))
                invalids[-1][-1] = invalids[-1][0]
        return Values(invalids)
    
    def _valid_length_range(self):
        if self.min_items is not None and self.max_items is not None:
            return self.min_items, self.max_items
        elif self.min_items is not None:
            return self.min_items, self.min_items + 10
        elif self.max_items is not None:
            return 0, self.max_items
        else:
            return 0, 10
    
    def __str__(self):
        return "Array: {}".format(str(self.datatype))

    def _equals(self, spec, res):
        if len(spec) != len(res):
            raise ValueError("Arrays have different lengths.")
        for v1, v2 in zip(spec, res):
            self.datatype.equals(v1, v2)
        return True


def test_array():
    _test_valid_args(Array, [
        [String()], [String(), 3],
        [String(), 5, 8], [String(), 5, 8, True],
        [Integer(1, 3), 2, 2, True]
    ])

    _test_invalid_args(Array, [
        [String(), -1], [String(), 3, -2],
        [String(), -2, -1], [String(), 6, 3],
        [Integer(1, 10), 11, 11, True]
    ])

    assert len(Array(Integer(), 3, 3).valids()[0]) == 3
    assert len(Array(Integer(), 10, 10).valids()[0]) == 10
    for i in range(10):
        assert 4 < len(Array(String(), 5, 9).valids()[0]) < 10
    for i in range(10):
        assert len(set(Array(Integer(1, 100), 10, 10, True).valids()[0])) == 10
    for unique_items in [True, False]:
        a = Array(Integer(), 5, 10, unique_items)
        for v in a.valids(10):
            assert a.valid(v, True)
        for v in a.invalids(1, True):
            with pytest.raises(ValueError):
                a.valid(v, True)



from enum import Enum as Enum_
import exrex
import pytest
import random
import re


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Value, Values


class StringPattern(Enum_):
    CHARS: "[A-Za-z]{8,16}"
    CHARS_NUMBERS: "[A-Za-z0-9]{8,16}"
    CHARS_NUMBERS_SPACES: "[A-Za-z0-9 ]{8,16}"
    CHAR_CHARS_NUMBERS_CHAR: "[A-Za-z][A-Za-z0-9]{6,14}[A-Za-z]"
    CHAR_CHARS_SPACES_CHAR: "[A-Za-z][A-Za-z ]{6,14}[A-Za-z]"
    CHAR_CHARS_NUMBERS_SPACES_CHAR: "[A-Za-z][A-Za-z0-9 ]{6,14}[A-Za-z]"

class String(Dummy):
    def __init__(self, min_length=None, max_length=None, pattern=None, invalid_pattern=None):
        super().__init__()
        if min_length and min_length < 0:
            raise ValueError("Min length cannot be smaller than 0 but is {}.".format(min_length))
        if max_length and max_length < 0:
            raise ValueError("Max length cannot be smaller than 0 but is {}.".format(max_length))
        if min_length and max_length and min_length > max_length:
            raise ValueError("Min length cannot be larger than max length.")
        if invalid_pattern and not pattern:
            raise ValueError("Must specify pattern with invalid_pattern")
        
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = pattern
        self.pattern_ = re.compile(pattern) if pattern else None
        self.invalid_pattern = invalid_pattern
    
    def valid(self, value, raise_error=False):
        if not isinstance(value, str):
            if raise_error:
                raise ValueError("{} is not a string but {}".format(value, type(value)))
            return False
        if self.min_length and len(value) < self.min_length:
            if raise_error:
                raise ValueError("{} < {}".format(value, self.min_length))
            return False
        if self.max_length and len(value) > self.max_length:
            if raise_error:
                raise ValueError("{} > {}".format(value, self.max_length))
            return False
        if self.pattern_ and not self.pattern_.match(value):
            if raise_error:
                raise ValueError("{} does not match the pattern {}".format(value, self.pattern))
        return True

    def _valids(self):
        if self.pattern:
            return Value(exrex.getone(self.pattern))
        if self.min_length is None and self.max_length is None:
            length = random.randint(8, 16)
        elif self.max_length is None:
            length = random.randint(self.min_length,
                                    2 * self.min_length)
        elif self.min_length is None:
            length = random.randint(math.floor(self.max_length / 2),
                                    self.max_length)
        else:
            length = random.randint(self.min_length, self.max_length)
        return Value(exrex.getone("[A-Za-z0-9]{{{}}}".format(length)))

    def _invalids(self):
        invalids = [
            random.randint(0, 100),
            random.random()
        ]
        if self.invalid_pattern:
            invalids.append(exrex.getone(self.invalid_pattern))
        return Values(invalids)

    def __str__(self):
        return "String: {} ({}, {})".format(self.pattern, self.min_length, self.max_length)


def test_string():
    _test_valid_args(String, [
        [1, 10], [20, 20],
        {"pattern": "[a-z]{10,20}"},
        {"pattern": "[a-z]{10,20}", "invalid_pattern": "[a-z]{9}"}
    ])

    _test_invalid_args(String, [
        [-3, 10], [10, 9],
        {"invalid_pattern": "[a-z]{9}"}
    ])
    
    for x in range(0,30):
        assert len(String(x, x).valids()[0]) == x
    for a in range(20):
        for b in range(a, 30):
            assert a <= len(String(a, b).valids()[0]) <= b
    
    for i in range(10):
        assert String(i, i, "0123456789").valids()[0] == "0123456789"
    assert len(String(8, 8, "[0-9]{4}").valids()[0]) == 4
    assert len(String(1, 1, "abcd[0-9]{4}").valids()[0]) == 8
    s = String(pattern="aaa", invalid_pattern="abc")
    assert "aaa" in s.valids(1, True)
    assert "abc" in s.invalids(1, True)



from faker import Faker as Faker_
import pytest


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Value, Values


class Faker(Dummy):
    BASIC_FAKER = Faker_()

    def __init__(self, kind, locales=None, invalids=None, args=None, kwargs=None):
        super().__init__()
        self.kind = kind
        self.locales = locales if locales else []
        self.args = args if args else []
        self.kwargs = kwargs if kwargs else {}
        if self.locales and len(self.locales) > 0:
            self.faker = Faker_(self.locales)
        else:
            self.faker = Faker.BASIC_FAKER
        if not hasattr(self.faker, kind):
            raise ValueError("faker does not have a property {}".format(kind))
        self._invalids_ = invalids if invalids else []
    
    def _valids(self):
        func = getattr(self.faker, self.kind)
        return Value(func(*self.args, **self.kwargs))
    
    def _invalids(self):
        return Values(self._invalids_)
    
    def __str__(self):
        return "Faker of {}".format(self.kind)


def test_faker():
    _test_valid_args(Faker, [
        ["url"], ["name"],
        ["first_name", "de_DE"],
        ["last_name", ["de_DE", "en_US"]],
        ["name", None, [1, 1.3, None]],
        ["name", None, lambda: 123],
    ])
    _test_invalid_args(Faker, [
        [], ["qwe"],
    ])
    
    assert len(Faker("name").valids(5)) == 5
    assert len(Faker("name").invalids(5)) == 0
    assert len(Faker("name", invalids=[1, 2, 3]).invalids(5)) == 3
    assert len(Faker("name", invalids=lambda: 123).invalids(6)) == 6
    assert len(Faker("date", args=["%Y-%m-%dT%H:%M:%SZ"]).valids(7)) == 7
    assert len(Faker("date", args=["%Y-%m-%dT%H:%M:%SZ"]).valids(8)) == 8
    
    #for i in range(10):
    #    o = Object(**props)
    #    for v in o.valids():
    #        assert o.valid(v, True)
    #for i in range(10):
    #    o = Object(**props)
    #    for v in o.invalids(1, True):
    #        with pytest.raises(ValueError):
    #            print(v)
    #            o.valid(v, True)



import pytest

def _test_valid_args(constr, argss):
    for args in argss:
        if isinstance(args, dict):
            print("VALID:", args.keys())
            x = constr(**args)
        elif isinstance(args, list):
            print("VALID:", [str(x) for x in args])
            x = constr(*args)
        else:
            raise ValueError("Invalid args type: {}".format(type(args)))
        x.valids()
        x.invalids()
        str(x)


def _test_invalid_args(constr, argss):
    for args in argss:
        if isinstance(args, dict):
            print("INVALID:", args.keys())
            with pytest.raises(Exception):
                x = constr(*args)
        elif isinstance(args, list):
            print("INVALID:", [str(x) for x in args])
            with pytest.raises(Exception):
                x = constr(**args)
        else:
            raise ValueError("Invalid args type: {}".format(type(args)))


class Value():
    def __init__(self, value):
        self.value = value


class Values():
    def __init__(self, values):
        self.values = values



import exrex
import math
import pytest
import random


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Value, Values


class Integer(Dummy):
    def __init__(self, minimum=None, maximum=None, multiple_of=None):
        super().__init__()
        if minimum is not None and maximum is not None and minimum > maximum:
            raise ValueError("minimum is greater than maximum")
        
        self.minimum = minimum
        self.maximum = maximum
        if multiple_of is not None:
            if minimum % multiple_of != 0:
                if minimum % multiple_of != 0:
                    if math.floor(minimum / multiple_of) == math.floor(maximum / multiple_of):
                        raise ValueError(
                            "Invalid multiple_of for minimum / maximum")
        self.multiple_of = multiple_of
        
    
    def valid(self, value, raise_error=False):
        if not isinstance(value, int):
            if raise_error:
                raise ValueError("{} is not an int but {}".format(value, type(value)))
            return False
        if self.minimum and value < self.minimum:
            if raise_error:
                raise ValueError("{} < {}".format(value, self.minimum))
            return False
        if self.maximum and value > self.maximum:
            if raise_error:
                raise ValueError("{} > {}".format(value, self.maximum))
            return False
        if self.multiple_of and (value % self.multiple_of) != 0:
            if raise_error:
                raise ValueError("{} is not multiple of {}".format(value, self.multiple_of))
            return False
        return True

    def _valids(self):
        if self.minimum is not None and self.maximum is not None:
            start = self.minimum
            end = self.maximum
        elif self.minimum is not None:
            start = self.minimum
            end = self.minimum + 10
        elif self.maximum is not None:
            start = self.maximum - 10
            end = self.maximum
        else:
            start = 1
            end = 100
        if self.multiple_of is None:
            r = Value(random.randint(start, end))
            return r
        else:
            a = int(math.ceil(start / self.multiple_of)) * self.multiple_of
            b = int(math.floor(end / self.multiple_of)) * self.multiple_of
            r = Value(random.randrange(a, b+1, self.multiple_of))
            return r

    def _invalids(self):
        invalids = [4.2, exrex.getone("[a-z]{5}")]
        if self.minimum is not None:
            invalids.append(self.minimum - 1)
        if self.maximum is not None:
            invalids.append(self.maximum + 1)
        return Values(invalids)

    def __str__(self):
        return "Number: [{}, {}]".format(self.minimum, self.maximum)


def test_integer():
    _test_valid_args(Integer, [
        [1, 10, 2], [-20, -10],
        [-16, -14, 5], [-16, -14, 8]
    ])

    _test_invalid_args(Integer, [
        [10, 1], [-10, -11],
        [1, 10, 100], [5, 10, 12],
        [-15, -10, 8]
    ])

    for count in range(10):
        assert len(Integer().valids(count, True)) == count
        assert len(Integer().valids(count, False)) == count
    assert Integer(1, 5, 3).valids() == [3]
    assert Integer(10, 15, 4).valids() == [12]
    assert Integer(-4, -4).valids() == [-4]
    for i in range(20):
        assert Integer(1, 6, 3).valids()[0] in [3, 6]
    i = Integer(10, 20, 2)
    for v in list(range(10, 22, 2)) + i.valids(10):
        assert i.valid(v, True)
    for v in [-10, 9, 21, 234, 11, 0.4] + i.invalids(1, True):
        with pytest.raises(ValueError):
            i.valid(v, True)



from .array import test_array
from .boolean import test_boolean
from .enum import test_enum
from .faker import test_faker
from .integer import test_integer
from .number import test_number
from .object import test_object
from .string import test_string



from .array import Array
from .boolean import Boolean
from .enum import Enum
from .faker import Faker
from .integer import Integer
from .number import Number
from .object import Object
from .string import String



import pytest


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Value, Values


class Enum(Dummy):
    def __init__(self, *values, invalids=None):
        super().__init__()
        if len(values) == 0:
            raise ValueError("Must provide at least one value for enum values.")
        if len(values) != len(set(values)):
            raise ValueError("Enum values must be unique.")
        self._valids_ = list(values)
        self._invalids_ = invalids if invalids else []
    
    def _valids(self):
        return Values(self._valids_)
    
    def _invalids(self):
        return Values(self._invalids_)
    
    def valid(self, value, raise_error=False):
        if value not in self._valids_:
            if raise_error:
                raise ValueError("{} not a valid enum value, must be one of {}".format(
                    value, self._valids_
                ))
                return False
        return True
    
    def __str__(self):
        return "Enum: {}".format(self._valids_)


def test_enum():
    _test_valid_args(Enum, [
        [1], [1, "2", 33]
    ])

    _test_invalid_args(Enum, [
        [], [1, 2, 1], [3, 3]
    ])

    assert Enum(1, 2, 3).valids(3) == [1, 2, 3]
    assert Enum(1, 2, invalids=[3, 4]).invalids(2) == [3, 4]
    for i in range(1,10):
        assert Enum(*list(range(1,i+1))).valid(i, True), i
    for i in range(1,10):
        with pytest.raises(ValueError):
            Enum(*list(range(i+1,i+5))).valid(i, True), i



import exrex
import pytest
import re


from .dummy import Dummy
from .string import String
from .util import _test_valid_args, _test_invalid_args, Value, Values


class Object(Dummy):
    def __init__(self,
                 props,
                 required=None,
                 no_additional_props=False):
        super().__init__()
        self.props = props
        if len(props) == 0:
            raise ValueError("Must specify at least on prop.")
        self.required = required if required else []
        self.no_additional_props = no_additional_props
    
    def valid(self, value, raise_error=False):
        if self.no_additional_props:
            for k in value:
                if k not in self.props:
                    if raise_error:
                        raise ValueError("unknown property: {}".format(k))
                    return False
        for k in self.required:
            if k not in value:
                if raise_error:
                    raise ValueError("property {} is required but  isssing".format(k))
                return False
        for k,v in value.items():
            if not self.props[k].valid(v, raise_error):
                return False
        return True
    
    def _valids(self):
        return Value({
            k: v.valids(1, False)[0]
            for k,v in self.props.items()
        })
    
    def _invalids(self, invalid_props=False):
        invalids = []
        for k in self.required:
            invalids.append(self.valids()[0])
            del invalids[-1][k]
        if invalid_props:
            for k,p in self.props.items():
                inv = self.props[k].invalids(1, False)
                if len(inv) > 0:
                    invalids.append(self.valids()[0])
                    invalids[-1][k] = inv[0]
        if self.no_additional_props:
            invalids.append(self.valids()[0])
            k = exrex.getone("[a-zA-Z0-9]{16,32}")
            invalids[-1][k] = k
        return Values(invalids)
    
    def __str__(self):
        return "Object: {}".format(list(self.props.keys()))

    def _equals(self, spec, res):
        pattern = re.compile(r'(?<!^)(?=[A-Z])')
        def to_key(original):
            return pattern.sub('_', original).lower()
        spec = spec if isinstance(spec, dict) else spec.to_dict()
        res = res if isinstance(res, dict) else res.to_dict()
        spec = {to_key(k):v for k,v in spec.items()}
        res = {to_key(k):v for k,v in spec.items()}
        props = {to_key(k):v for k,v in self.props.items()}
        for k, v in spec.items():
            if k not in props:
                raise KeyError(
                    "Key {} not present in Object's definition.".format(k))
            if k not in res:
                raise KeyError(
                    "Key {} not present in result.".format(k))
            if not props[k].equals(v, res[k]):
                raise ValueError("Properties for key {} do not match ({} != {})".format(k, v, res[k]))
        return True


def test_object():
    _test_valid_args(Object, [
        [{"a": String(), "b": String()}]
    ])
    _test_invalid_args(Object, [
        [], [{}]
    ])

    o = Object({
        "a": String(pattern="aaa"),
        "b": String(pattern="bbb"),
    })
    assert o.valids()[0] == {"a": "aaa", "b": "bbb"}
    props = {
        "a": String(pattern="[a-z]{5}", invalid_pattern="aaa"),
        "b": String(pattern="[A-Z]{10}", invalid_pattern="bbb"),
    }
    flags = {
        "invalid_props": True,
    }
    for i in range(10):
        o = Object(props, no_additional_props=False)
        for v in o.valids():
            print(v)
            assert o.valid(v, True)
    for i in range(10):
        o = Object(props, no_additional_props=False)
        for v in o.invalids(1, True, **flags):
            with pytest.raises(ValueError):
                print(v)
                o.valid(v, True)



import random


from .util import Value, Values


class Dummy:
    def _generate(generator, count, exhaustive, **kwargs):
        res = generator(**kwargs)
        if isinstance(res, Value):
            return [generator(**kwargs).value for _ in range(count)]
        elif isinstance(res, Values):
            if isinstance(res.values, (list, tuple)):
                if exhaustive or count >= len(res.values):
                    return [x for x in res.values]
                else:
                    return list(random.sample(res.values, count))
            else:
                return [res.values(**kwargs) for _ in range(count)]
        else:
            raise TypeError("Result of generator has invalid type: {}".format(type(res)))
    
    def valids(self, count=1, exhaustive=False, **kwargs):
        return Dummy._generate(self._valids, count, exhaustive, **kwargs)
    
    def invalids(self, count=1, exhaustive=False, **kwargs):
        return Dummy._generate(self._invalids, count, exhaustive, **kwargs)

    def _valids(self, **kwargs):
        raise NotImplementedError("_valids() it not implemented for {}.".format(type(self)))

    def _invalids(self, **kwargs):
        raise NotImplementedError("_invalids() it not implemented for {}.".format(type(self)))

    def valid(self, value, raise_error=False):
        raise NotImplementedError("valid() is not implemented for {}.".format(type(self)))

    def examples(self):
        return {
            "valid": self.valids(),
            "invalid": self.invalids(),
        }
    
    def equals(self, v1, v2):
        if v1 is None and v2 is None:
            return True
        if v1 is None or v2 is None:
            return False
        return self._equals(v1, v2)

    def _equals(self, v1, v2):
        return v1 == v2



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="evonik-dummy",
    version="0.0.3",
    author="Benjamin Schiller",
    author_email="benjamin.schiller@evonik.com",
    description="Generate dummy data, e.g., for testing APIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://evodl.visualstudio.com/evonik-dummy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "exrex>=0.10.5",
        "faker>=4.1.1",
        "pytest>=6.0.1"
    ],
)



import exrex
import math
import pytest
import random
import sys


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Value, Values



class Number(Dummy):
    def __init__(self, minimum=None, maximum=None, multiple_of=None,
                 exclusive_minimum=False, exclusive_maximum=False):
        super().__init__()
        if minimum is not None and maximum is not None and minimum > maximum:
            raise ValueError("minimum is greater than maximum")
        if exclusive_minimum or exclusive_maximum:
            if minimum is not None and maximum is not None and minimum == maximum:
                raise ValueError(
                    "minimum and maximum cannot be equal and excluded")
        
        self.minimum = minimum
        self.maximum = maximum
        if multiple_of is not None:
            if minimum is not None and minimum % multiple_of != 0:
                if maximum is not None and maximum % multiple_of != 0:
                    if math.floor(minimum / multiple_of) == math.floor(maximum / multiple_of):
                        raise ValueError(
                            "Invalid multiple_of for minimum / maximum")
        self.multiple_of = multiple_of
        self.exclusive_minimum = exclusive_minimum
        self.exclusive_maximum = exclusive_maximum
    
    def valid(self, value, raise_error=False):
        if not isinstance(value, float) and not isinstance(value, int):
            if raise_error:
                raise ValueError("{} is not a float but {}".format(value, type(value)))
            return False
        if self.minimum and (
            (self.exclusive_minimum and value <= self.minimum) or
            (not self.exclusive_minimum and value < self.minimum)
        ):
            if raise_error:
                raise ValueError("{} < {}".format(value, self.minimum))
            return False
        if self.maximum and (
            (self.exclusive_maximum and value >= self.maximum) or
            (not self.exclusive_maximum and value > self.maximum)
        ):
            if raise_error:
                raise ValueError("{} > {}".format(value, self.maximum))
            return False
        if self.multiple_of and (
            (value % self.multiple_of) > sys.float_info.epsilon
            and
            (value % self.multiple_of) - self.multiple_of > sys.float_info.epsilon
        ) != 0:
            if raise_error:
                # print("v, m:", value, self.multiple_of)
                # print(value % self.multiple_of)
                # print((value % self.multiple_of) - self.multiple_of)
                # print((value % self.multiple_of) - self.multiple_of < sys.float_info.epsilon)
                raise ValueError("{} is not multiple of {}".format(value, self.multiple_of))
            return False
        return True

    def _valids(self):
        if self.minimum is not None and self.maximum is not None:
            start = self.minimum
            end = self.maximum
        elif self.minimum is not None:
            start = self.minimum
            end = self.minimum + 10
        elif self.maximum is not None:
            start = self.maximum - 10
            end = self.maximum
        elif self.multiple_of is not None:
            start = min(10 * self.multiple_of, 0.0)
            end = max(10 * self.multiple_of, 0.0)
        else:
            start = 0.0
            end = 1.0

        counter = 0
        while True:
            if counter > 100:
                raise StopIteration(
                    "Hit minimum or maximum more than 100 times.")
            
            if self.multiple_of is None:
                v = random.uniform(start, end)
            else:
                start_ = int(math.ceil(start / self.multiple_of))
                end_ = int(math.floor(end / self.multiple_of))
                v = random.randint(start_, end_) * self.multiple_of
            
            if self.exclusive_minimum and v == start:
                counter += 1
                continue
            if self.exclusive_maximum and v == end:
                counter += 1
                continue
            
            return Value(v)

    def _invalids(self):
        invalids = [4.2, exrex.getone("[a-z]{5}")]
        if self.minimum is not None:
            invalids.append(self.minimum - 1.0)
        if self.maximum is not None:
            invalids.append(self.maximum + 1)
        return Values(invalids)

    def __str__(self):
        l = "(" if self.exclusive_minimum else "["
        r = ")" if self.exclusive_maximum else "]"
        return "Number: {}{}, {}{}".format(l, self.minimum, self.maximum, r)


def test_number():
    _test_valid_args(Number, [
        [1.1, 10.3, 2.5], [-20, -10.0],
        [-16.2, -14, 0.4], [-17.6, -14, 8.8]
    ])

    _test_invalid_args(Number, [
        [10, 1], [-10, -11],
        [1, 10, 100], [5, 10, 12],
        [-15, -10, 8]
    ])

    for count in range(10):
        assert len(Number().valids(count, True)) == count
        assert len(Number().valids(count, False)) == count
    
    assert Number(3, 3).valids() == [3]
    assert Number(14.1, 15.2, 1.51).valids() == [15.1]
    for i in range(20):
        assert Number(1, 1.6, 0.3).valids()[0] in [1.2, 1.5]
        assert Number(-4, -2, 1).valids()[0] in [-4, -3, -2]
    
    i = Number(10, 12, 0.4)
    for v in list(range(6)):
        v = 10.0 + v * 0.4
    for v in i.valids(10):
        assert i.valid(v, True)
    for v in [-10, 9.9999, 12.0001, 234.2] + i.invalids(1, True):
        with pytest.raises(ValueError):
            i.valid(v, True)
    i = Number(10, 12, 0.4, True, True)
    for v in [10, 12]:
        with pytest.raises(ValueError):
            i.valid(v, True)



import pytest


from .dummy import Dummy
from .util import _test_valid_args, _test_invalid_args, Values


class Boolean(Dummy):
    def __init__(self, *invalids):
        super().__init__()
        if len(invalids) != len(set(invalids)):
            raise ValueError("invalids must be unique.")
        for invalid in invalids:
            if not isinstance(invalid, bool):
                raise ValueError("{} is not a boolean but a {}.".format(
                    value, type(value)
                ))
        if len(invalids) == 0:
            valids = [True, False]
        elif len(invalids) == 1:
            valids = [not invalids[0]]
        else:
            valids = []
        self._valids_ = valids
        self._invalids_ = invalids

    def _valids(self):
        return Values(self._valids_)

    def _invalids(self):
        return Values(self._invalids_)

    def valid(self, value, raise_error=False):
        if not isinstance(value, bool):
            if raise_error:
                raise ValueError("{} is not of type bool but {}".format(
                    value, type(value)
                ))
            return False
        if value not in self._valids_:
            if raise_error:
                raise ValueError("{} not a valid value, must be in {}".format(
                    value, self._valids_
                ))
            return False
        return True
    
    def __str__(self):
        return "Boolean: {} / {}".format(self._valids_, self._invalids_)


def test_boolean():
    _test_valid_args(Boolean, [
        [], [True], [False],
        [True, False], [False, True]
    ])

    _test_invalid_args(Boolean, [
        [False, False], [True, True],
        [True, False, True], [False, True, False],
        [True for i in range(3)]
    ])

    assert Boolean(False).valids() == [True]
    assert Boolean(False).invalids() == [False]
    assert Boolean(True).valids() == [False]
    assert Boolean(True).invalids() == [True]
    assert Boolean().valids(2) == [True, False]
    assert Boolean().invalids() == []
